/*
 * Copyright (C) 2003-2007 eXo Platform SAS.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see<http://www.gnu.org/licenses/>.
 */
package org.exoplatform.services.jcr.impl.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.jcr.PathNotFoundException;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.ValueFormatException;

import org.exoplatform.services.jcr.dataflow.DataManager;
import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.core.value.ValueFactoryImpl;

/**
 * Created by The eXo Platform SAS
 * 15.05.2006
 * 
 * NodeData bulk reader.
 * 
 * @author <a href="mailto:peter.nedonosko@exoplatform.com.ua">Peter Nedonosko</a>
 * @version $Id: NodeDataReader.java 12841 2007-02-16 08:58:38Z peterit $
 */
public class NodeDataReader extends ItemDataReader {

  private final HashMap<InternalQName, NodeInfo> nodes = new HashMap<InternalQName, NodeInfo>();
  private final HashMap<InternalQName, NodeInfo> nodesByType = new HashMap<InternalQName, NodeInfo>();
  
  private PropertyDataReader nodePropertyReader = null;
  
  private final List<NodeData> skiped = new ArrayList<NodeData>();
  private boolean rememberSkiped = false;
  
  private class NodeInfo {
    private final InternalQName nodeName;
    private final List<NodeDataReader> childNodesReaders;

    NodeInfo(InternalQName nodeName, List<NodeDataReader> childNodesReaders) {
      this.nodeName = nodeName;
      this.childNodesReaders = childNodesReaders;
    }
    
    public InternalQName getNodeName() {
      return nodeName;
    }

    public List<NodeDataReader> getChildNodesReaders() {
      return childNodesReaders;
    }
  }
  
  public NodeDataReader(NodeData node, DataManager dataManager, ValueFactoryImpl valueFactory) {
    super(node, dataManager, valueFactory);
  }
  
  public NodeDataReader forNodesByType(InternalQName name) {
    nodesByType.put(name, new NodeInfo(name, new ArrayList<NodeDataReader>()));
    return this;
  }
  
  public NodeDataReader forNode(InternalQName name) {
    nodes.put(name, new NodeInfo(name, new ArrayList<NodeDataReader>()));
    return this;
  }
  
  public List<NodeDataReader> getNodes(InternalQName name) throws PathNotFoundException {
    List<NodeDataReader> nr = nodes.get(name).getChildNodesReaders();
    if (nr.size()>0)
      return nr;
    throw new PathNotFoundException("Node with name " + parent.getQPath().getAsString() 
        + name.getAsString() + " not found");
  }
  
  public List<NodeDataReader> getNodesByType(InternalQName typeName) throws PathNotFoundException {
    List<NodeDataReader> nr = nodesByType.get(typeName).getChildNodesReaders();
    if (nr.size()>0)
      return nr;
    throw new PathNotFoundException("Nodes with type " + typeName.getAsString() + " not found. Parent "
        + parent.getQPath().getAsString());
  }
  
  public Value getPropertyValue(InternalQName name) throws ValueFormatException, PathNotFoundException, RepositoryException {
    return nodePropertyReader.getPropertyValue(name);
  }
  
  public ValueData getPropertyValueData(InternalQName name) throws ValueFormatException, PathNotFoundException, RepositoryException {
    return nodePropertyReader.getPropertyValueData(name);
  }
  
  public List<Value> getPropertyValues(InternalQName name) throws ValueFormatException, PathNotFoundException, RepositoryException {
    return nodePropertyReader.getPropertyValues(name);
  }
  
  public List<ValueData> getPropertyValueDatas(InternalQName name) throws ValueFormatException, PathNotFoundException, RepositoryException {
    return nodePropertyReader.getPropertyValueDatas(name);
  }
  
  /**
   * Read node properties
   * @param name
   * @param type
   * @param multiValued
   * @return
   */
  public PropertyDataReader forProperty(InternalQName name, int type) {
    if (nodePropertyReader == null) {
      nodePropertyReader = new PropertyDataReader(parent, dataManager, valueFactory);
    } 
    return nodePropertyReader.forProperty(name, type);
  }
  
  public void read() throws RepositoryException {

    cleanReaders();

    if (nodePropertyReader != null) {
      nodePropertyReader.read();
    }

    if (nodes.size() > 0 || nodesByType.size() > 0 || rememberSkiped) {
      
      List<NodeData> ndNodes = dataManager.getChildNodesData(parent);
      for (NodeData node : ndNodes) {

        boolean isSkiped = true;

        NodeDataReader cnReader = new NodeDataReader(node, dataManager, valueFactory);

        NodeInfo nodeInfo = nodes.get(node.getQPath().getName());
        if (nodeInfo != null) {
          nodeInfo.getChildNodesReaders().add(cnReader);
          isSkiped = false;
        }

        NodeInfo nodesByTypeInfo = nodesByType.get(node.getPrimaryTypeName());
        if (nodesByTypeInfo != null) {
          nodesByTypeInfo.getChildNodesReaders().add(cnReader);
          isSkiped = false;
        }

        if (isSkiped && rememberSkiped) {
          skiped.add(node);
        }
      }
    }
  }
  
  private void cleanReaders() {
    for (Map.Entry<InternalQName, NodeInfo> nodesEntry: nodes.entrySet()) {
      nodesEntry.getValue().getChildNodesReaders().clear();
    }
    for (Map.Entry<InternalQName, NodeInfo> nodesEntry: nodesByType.entrySet()) {
      nodesEntry.getValue().getChildNodesReaders().clear();
    }
    skiped.clear();
  }
  
  public void clean() {
    nodes.clear();
    nodesByType.clear();
    skiped.clear();
    nodePropertyReader = null;
  }

  public boolean isRememberSkiped() {
    return rememberSkiped;
  }

  public void setRememberSkiped(boolean rememberSkiped) {
    this.rememberSkiped = rememberSkiped;
  }

  public List<NodeData> getSkiped() {
    return skiped;
  }
}
