/*
 * Copyright (C) 2003-2007 eXo Platform SAS.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see<http://www.gnu.org/licenses/>.
 */
package org.exoplatform.services.jcr.impl.util.io;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.CountDownLatch;

/**
 * Created by The eXo Platform SAS
 * Author : Peter Nedonosko
 *          peter.nedonosko@exoplatform.com.ua
 * 05.10.2007  
 *
 * For use in persistent layer, i.e. JDBCStorageConnection (ClenableFileValueData), 
 * will be shared across the Workspace cache with many Sessions . 
 * Swap files creation (like in JDBCStorageConnection.readValueData(String, int, int)) managed by SwapFile.get(File, String) method. 
 * There are no way to get swap file in another way.
 *  
 * A SwapFile extends the SpoolFile. But in runtime all swap files will be stored in global map used for 
 * prevent files rewriting in concurrent environment (case of issue JCR-329). 
 * Till the SwapFile spool operation in progress all other users who will attempt to get the file will wait the 
 * operation completion (java.util.concurrent.CountDownLatch used).
 *
 * @author <a href="mailto:peter.nedonosko@exoplatform.com.ua">Peter Nedonosko</a>
 * @version $Id: SwapFile.java 8440 2007-11-30 15:52:29Z svm $
 */
public class SwapFile extends SpoolFile {
  
  protected static Map<String, SwapFile> inShare = new HashMap<String, SwapFile>();  
  
  protected CountDownLatch spoolLatch = null;
  
  private SwapFile(File parent, String child) {
    super(parent, child);
  }
  
  /**
   * Obtain SwapFile by parent file and name.
   * 
   * If the file was swapped before and still in use it will be returned, 
   * i.e. same object of java.io.File will be returned.
   * 
   * if the file swapping (writing) now at this time the caller thread 
   * will wait till the swap process will be finished.    
   * 
   * @param parent - File 
   * @param child - String
   * @return SwapFile
   * @throws IOException
   */
  public static SwapFile get(final File parent, final String child) throws IOException {
    synchronized (inShare) {
      SwapFile newsf = new SwapFile(parent, child);
      String absPath = newsf.getAbsolutePath();
      
      SwapFile swapped = inShare.get(absPath);
      if (swapped != null) {
        CountDownLatch spoolLatch = swapped.spoolLatch;
        if (spoolLatch != null)
          try {
            spoolLatch.await(); // wait till the file will be done
          } catch (final InterruptedException e) {
            // thinking that is ok, i.e. this thread is interrupted 
            throw new IOException("Swap file read error " + swapped.getAbsolutePath() + ". " + e) {
              @Override
              public Throwable getCause() {
                return e;
              }
            };
          }  
        return swapped; 
      }
      
      newsf.spoolLatch = new CountDownLatch(1);
      inShare.put(absPath, newsf);
      return newsf;
    }
  }

  /** 
   * Tell if the file already spooled - ready for use.
   * @return boolean flag
   */
  public boolean isSpooled() {
    return spoolLatch == null;
  }
  
  /**
   * Mark the file ready for read. 
   */
  public void spoolDone() {
    final CountDownLatch sl = this.spoolLatch;
    this.spoolLatch = null;
    sl.countDown();
  }
  
  //------ java.io.File ------
  
  /**
   * Delete file if it was not used by any other thread.  
   */
  @Override
  public boolean delete() {
    synchronized (inShare) {
      if (super.delete()) {
        // remove from shared files list
        inShare.remove(getAbsolutePath());
                
        // make sure that the file doesn't make any other thread await in 'get' method
        // impossible case as 'delete' and 'get' which may waiting for, synchronized by inShare map. 
        //spoolDone();
        
        return true;
      }
      
      return false;
    }
  }
  
  /**
   * Not applicable. Call get(File, String) method instead
   * 
   * @throws IOException
   */
  public static SwapFile createTempFile(String prefix, String suffix,
      File directory) throws IOException {
    throw new IOException("Not applicable. Call get(File, String) method instead");
  }
}
 